#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    Algorithmes de tris.
"""

import random

__author__ = "Nicolas Pronost, Laure Gonnord et Olivier Laurent"
__copyright__ = "Univ Lyon1, 2019"


#  Test de tri

def est_trie(tab):
    """teste si tab est trié"""
    for i in range(len(tab)-1):
        if tab[i] > tab[i+1]:
            return False
    return True


# Tri par insertion

def tri_par_insertion(tab):
    """trie le "tableau" tab par l'algorithme tri insertion"""
    # On place successivement chaque élément à sa place
    # parmi les éléments d'indices plus petits déjà triés.
    for j in range(1, len(tab)):
        # On commence par mémoriser l'élément en cours de placement.
        elt_a_placer = tab[j]
        # On parcourt tab de la droite vers la gauche
        # à partir de l'élément précédent.
        i = j - 1
        while i >= 0 and tab[i] > elt_a_placer:
            # On décale tab[i] vers la droite
            # si il est plus grand que elt_a_placer.
            tab[i+1] = tab[i]
            i -= 1
        tab[i+1] = elt_a_placer  # Placement de l'élément à la sortie


# Tri par sélection

def tri_par_selection(tab):
    """trie tab par l'algorithme de sélection"""
    for i in range(len(tab)-1):  # de 0 à n-2 (n la longueur de tab)
        indmin = i
        for j in range(i+1, len(tab)):  # de i+1 à n-1, recherche du min
            if tab[j] < tab[indmin]:
                indmin = j
        tab[indmin], tab[i] = tab[i], tab[indmin]  # échange de valeurs


# Tri rapide (quick sort)
# dans un premier temps on ne se soucie pas de faire les choses en place

def partitionner(tab, elt):
    """partitionne tab entre les éléments < elt et ceux >= elt
       rend une paire de listes"""
    gauche, droite = [], []  # initialisation
    for val in tab:
        if val < elt:
            gauche.append(val)  # on range dans gauche les éléments < elt
        else:
            droite.append(val)  # on range dans droite les éléments >= elt
    return gauche, droite


def tri_rapide_rec(tab):
    """trie tab par l'algorithme de tri rapide (quick sort)"""
    if len(tab) <= 1:                  # liste vide ou singleton
        return tab                     # dans ce cas tab est trié
    # choix de l'indice du pivot (au choix parmi les 3 lignes qui suivent)
    # ipivot = 0                              # comme premier élément de tab
    ipivot = len(tab)-1                     #  comme dernier élément de tab
    # ipivot = random.randint(0, len(tab)-1)  # aléatoirement
    pivot = tab[ipivot]  # le pivot
    gauche, droite = partitionner(tab[:ipivot] + tab[ipivot+1:], pivot)
    return tri_rapide_rec(gauche) + [pivot] + tri_rapide_rec(droite)


def tri_rapide1(tab):
    """transforme tab en sa version triée en utilisant tri_rapide_rec"""
    tab_sort = tri_rapide_rec(tab)
    for i in range(len(tab)):
        tab[i] = tab_sort[i]


## on fait maintenant les choses en place (économie de mémoire)

# plusieurs manières de partitionner sont possibles, en voici 2
def partitionner_en_place(tab, deb, fin, ipivot):
    """partitionne les éléments de tab entre deb et fin
       par rapport à pivot = tab[ipivot]
       ipivot est supposé entre deb et fin
       retourne l'indice ipart de tab[ipivot] après la partition
       les éléments de tab entre deb et ipart sont <= pivot
       les éléments de tab entre ipart+1 et fin sont > pivot"""
    tab[ipivot], tab[fin] = tab[fin], tab[ipivot]  # on stocke pivot à la fin
    j = deb            # entre deb et j -1, une suite d'éléménts <= pivot
    for i in range(deb, fin):  # entre j et i-1, une suite d'éléments > pivot
        if tab[i] <= tab[fin]: # si on a un élément <= pivot, on le place en j
            tab[i], tab[j] = tab[j], tab[i]
            j += 1
    tab[fin], tab[j] = tab[j], tab[fin]  # on replace le pivot
    return j


def partitionner_en_place2(tab, deb, fin, ipivot):
    """partitionne les éléments de tab entre deb et fin
       par rapport à pivot = tab[ipivot]
       ipivot est supposé entre deb et fin
       retourne l'indice ipart de tab[ipivot] après la partition
       les éléments de tab entre deb et ipart sont < pivot
       les éléments de tab entre ipart+1 et fin sont >= pivot"""
    tab[ipivot], tab[fin] = tab[fin], tab[ipivot]  # on stocke pivot à la fin
    i, j = deb, fin-1                              # initialisation
    while True:
        while tab[i] < tab[fin]:              # on cherche une valeur >= pivot
                                              # on s'arrête forcément si i = fin
            i += 1
        while j >= i and tab[j] >= tab[fin]:  # on cherche une valeur < pivot
                                              # on s'arrête forcément si j = i
            j -= 1
        if i < j:  # on a trouvé deux valeurs à permuter
            tab[i], tab[j] = tab[j], tab[i]
        else:     # on est arrivés au bout (i = fin ou i = j)
            tab[i], tab[i+1:fin+1] = tab[fin], tab[i:fin]  # on replace le pivot
            return i


def tri_rapide_rec_en_place(tab, deb, fin):
    """trie les éléments de tab entre deb et fin par tri rapide"""
    if deb < fin:  # sinon c'est déjà trié
        # choix de l'indice du pivot (au choix parmi les 3 lignes qui suivent)
        # ipivot = deb                       # comme premier élément
        ipivot = fin                       #  comme dernier élément
        # ipivot = random.randint(deb, fin)  # aléatoirement
        ipivot_part = partitionner_en_place(tab, deb, fin, ipivot)
        # ipivot_part = partitionner_en_place2(tab, deb, fin, ipivot)
        tri_rapide_rec_en_place(tab, deb, ipivot_part-1)
        tri_rapide_rec_en_place(tab, ipivot_part+1, fin)


def tri_rapide2(tab):
    """tri rapide de tab via tri_rapide_rec_en_place"""
    tri_rapide_rec_en_place(tab, 0, len(tab)-1)


# Tri par fusion
# dans un premier temps on ne se soucie pas de faire les choses en place
# on pourra choisir soit la fusion récursive soit la fusion itérative

def fusion_rec(gauche, droite):
    """Fusion récursive :
       construit la fusion triée des deux listes triées gauche et droite"""
    if len(gauche) == 0:  # si la liste gauche est vide, la fusion est facile
        return droite
    if len(droite) == 0:  # si la liste droite est vide, la fusion est facile
        return gauche
    if gauche[0] < droite[0]:  # on place en tête le plus petit élément
        return [gauche[0]] + fusion_rec(gauche[1:], droite)
    else:
        return [droite[0]] + fusion_rec(gauche, droite[1:])


def fusion_it(gauche, droite):
    """Fusion itérative :
       construit la fusion triée des deux listes triées gauche et droite"""
    igauche, idroite = 0, 0
    result = []
    while igauche < len(gauche) and idroite < len(droite):
        if gauche[igauche] < droite[idroite]:
            result.append(gauche[igauche])
            igauche += 1
        else:
            result.append(droite[idroite])
            idroite += 1
    # l'une des 2 listes est désormais vide, on ajoute ce qui reste
    result += gauche[igauche:]
    result += droite[idroite:]
    return result


def tri_fusion_rec(tab):
    """trie tab par l'algorithme de fusion
       version récursive"""
    if len(tab) <= 1:                  # liste vide ou singleton
        return tab                     # dans ce cas tab est trié
    milieu = len(tab) // 2             # on coupe en 2 au milieu
    gauche = tri_fusion_rec(tab[:milieu])  # trie la partie gauche de tab
    droite = tri_fusion_rec(tab[milieu:])  # trie la partie droite de tab
#    return fusion_it(gauche, droite)       # fusionne les deux
    return fusion_rec(gauche, droite)      # fusionne les deux


def tri_fusion1(tab):
    """transforme tab en sa version triée en utilisant tri_fusion_rec"""
    tab_sort = tri_fusion_rec(tab)
    for i in range(len(tab)):
        tab[i] = tab_sort[i]


# on fait maintenant les choses en place (économie de mémoire)
# on pourra choisir soit la fusion récursive soit la fusion itérative

# une fonction récursive de fusion en place
def fusion_rec_en_place(tab, beg, mid, end):
    """fusionne les parties de tab situées
       entre beg et mid-1 et entre mid et end"""
    if beg < mid and mid <= end:  # on teste qu'il y a encore des éléments
                                  # des 2 côtés sinon c'est fini
        if tab[beg] <= tab[mid]:  # on compare les 1ers éléments des 2 paquets
            fusion_rec_en_place(tab, beg+1, mid, end)
        else:  # si le plus petit est à droite, on le met à gauche et on décale
            tab[beg], tab[beg+1:mid+1] = tab[mid], tab[beg:mid]
            fusion_rec_en_place(tab, beg+1, mid+1, end)


# une fonction itérative de fusion en place
def fusion_it_en_place(tab, beg, mid, end):
    """fusionne les parties de tab situées
       entre beg et mid-1 et entre mid et end"""
    i, j = beg, mid
    while i < j and j <= end:  # on teste qu'il y a encore des éléments
                               # des 2 côtés sinon c'est fini
        if tab[i] > tab[j]: # si le plus petit est à droite,
                            # on le met à gauche et on décale
            tab[i], tab[i+1:j+1] = tab[j], tab[i:j]
            j += 1
        i += 1   # l'élément d'indice i est désormais à sa place


def tri_fusion_rec_en_place(tab, beg, end):
    """tri_fusion des élémentes de tab entre beg et end"""
    if beg < end:
        mid = (beg + end) // 2
        tri_fusion_rec_en_place(tab, beg, mid)   # trie la partie gauche de tab
        tri_fusion_rec_en_place(tab, mid+1, end) # trie la partie droite de tab
#        fusion_rec_en_place(tab, beg, mid+1, end) # fusionne les deux
        fusion_it_en_place(tab, beg, mid+1, end) # fusionne les deux


def tri_fusion2(tab):
    """tri_fusion de tab"""
    tri_fusion_rec_en_place(tab, 0, len(tab)-1)


